# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import base64
from json import dumps as json_dumps
import re
from StringIO import StringIO
from urlparse import urljoin

import requests
from requests.status_codes import codes as http_codes


__metaclass__ = type


def valid_fingerprint(fingerprint):
    """Is the fingerprint of valid form."""
    # Fingerprints of v3 keys are md5, fingerprints of v4 keys are sha1;
    # accordingly, fingerprints of v3 keys are 128 bit, those of v4 keys
    # 160. Check therefore for strings of hex characters that are 32
    # (4 * 32 == 128) or 40 characters long (4 * 40 = 160).
    if len(fingerprint) not in (32, 40):
        return False
    if re.match(r"^[\dA-F]+$", fingerprint) is None:
        return False
    return True


def sanitize_fingerprint(fingerprint):
    """Sanitize a GPG fingerprint."""
    # remove whitespaces, truncate to max of 40 (as per v4 keys) and
    # convert to upper case
    fingerprint = fingerprint.replace(' ', '')
    fingerprint = fingerprint[:40].upper()

    if not valid_fingerprint(fingerprint):
        return None

    return fingerprint


def sanitize_fingerprint_or_raise(fingerprint):
    """Check the sanity of 'fingerprint'.

    If 'fingerprint' is a valid fingerprint, the sanitised version will be
    returned (see sanitize_fingerprint).

    Otherwise, a ValueError will be raised.
    """
    sane_fingerprint = sanitize_fingerprint(fingerprint)
    if sane_fingerprint is None:
        raise ValueError("Invalid fingerprint: %r." % fingerprint)
    return sane_fingerprint


class GPGServiceException(Exception):
    """Raised when we get an error from the gpgservice.
    """


class GPGClient:

    def __init__(self, endpoint=None, timeout=None):
        """Create a new GPGClient instance.

        The `endpoint` and `timeout` parameters will, if supplied control where
        the client connects, and what timeout the http connections use. If not
        specified, then the user must override `get_endpoint` and `get_timeout`
        to supply these values instead.

        If `endpoint` and `timeout` are not supplied, *and* `get_endpoint` and
        `get_timeout` are not overridden, a TypeError will be raised.
        """
        self.write_hooks = set()
        self._endpoint = endpoint
        self._timeout = timeout
        if not (endpoint or self.get_endpoint()):
            raise TypeError(
                "Either specify `endpoint` or override `get_endpoint`")
        if not (timeout or self.get_timeout()):
            raise TypeError(
                "Either specify `timeout` or override `get_timeout`")

    # These are not properties since they get overridden in launchpad, and
    # overriding properties in subclasses makes it non-obvious where the values
    # come from.
    def get_endpoint(self):
        return self._endpoint

    def get_timeout(self):
        return self._timeout

    def getKeysForOwner(self, owner_id):
        """Get a list of keys for a given owner.

        :raises GPGServiceException: If we get an error from the gpgservice.
        :raises socket.error" on socket-level errors (connection timeouts etc)
        """
        path = '/users/%s/keys' % self._encode_owner_id(owner_id)
        resp = self._request('get', path)
        if resp.status_code != http_codes['OK']:
            self.raise_for_error(resp)
        return resp.json()

    def addKeyForOwner(self, owner_id, fingerprint):
        """Add a GPG key.

        :raises ValueError: if the fingerprint isn't valid.
        :raises GPGServiceException: If we get an error from the gpgservice.
        :raises socket.error" on socket-level errors (connection timeouts etc)

        :returns: The JSON representation of the key that was just added.
        """
        fingerprint = sanitize_fingerprint_or_raise(fingerprint)
        path = '/users/%s/keys' % self._encode_owner_id(owner_id)
        data = dict(fingerprint=fingerprint)
        resp = self._request('post', path, json=data)
        if resp.status_code == http_codes['CREATED']:
            self._notify_writes()
        elif resp.status_code != http_codes['OK']:
            self.raise_for_error(resp)
        return resp.json()

    def disableKeyForOwner(self, owner_id, fingerprint):
        """Disable a GPG key.

        :raises ValueError: if the fingerprint isn't valid.
        :raises GPGServiceException: If we get an error from the gpgservice.
        :raises socket.error" on socket-level errors (connection timeouts etc)
        """
        fingerprint = sanitize_fingerprint_or_raise(fingerprint)
        path = '/users/%s/keys/%s' % (self._encode_owner_id(owner_id),
                                      fingerprint)
        resp = self._request('delete', path)
        if resp.status_code == http_codes['OK']:
            self._notify_writes()
        else:
            self.raise_for_error(resp)

    def getKeyByFingerprint(self, fingerprint):
        """Get a GPG key by it's fingerprint.

        :raises ValueError: if the fingerprint isn't valid.
        """
        fingerprint = sanitize_fingerprint_or_raise(fingerprint)
        path = '/keys/%s' % fingerprint
        resp = self._request('get', path)
        if resp.status_code == http_codes['OK']:
            return resp.json()
        elif resp.status_code == http_codes['NOT_FOUND']:
            return None
        else:
            self.raise_for_error(resp)

    def getKeysByFingerprints(self, fingerprints):
        if len(fingerprints) == 1:
            return [self.getKeyByFingerprint(fingerprints[0])]
        fingerprint = ','.join(
            [sanitize_fingerprint_or_raise(f) for f in fingerprints])
        path = '/keys/%s' % fingerprint
        resp = self._request('get', path)
        if resp.status_code == http_codes['OK']:
            return resp.json()['keys']
        elif resp.status_code == http_codes['NOT_FOUND']:
            return []
        else:
            self.raise_for_error(resp)

    def registerWriteHook(self, hook_callable):
        """Register a write hook.

        The hook_callable will be called with no arguments whenever an
        operation is performed that modifies the GPG database.

        :raises TypeError: if hook_callable is not a callable.
        :raises GPGServiceException: If we get an error from the gpgservice.
        """
        if not callable(hook_callable):
            raise TypeError("'hook_callable' parameter must be a callable.")
        self.write_hooks.add(hook_callable)

    def unregisterWriteHook(self, hook_callable):
        """Deregister a write hook that was registered with register_write_hook.

        :raises ValueError: if hook_callable was not registered.
        """
        if hook_callable not in self.write_hooks:
            raise ValueError("%r not registered.")
        self.write_hooks.remove(hook_callable)

    def addKeyForTest(self, owner_id, keyid, fingerprint, keysize,
                      algorithm, enabled, can_encrypt):
        """Add a key to the gpgservice without checking the keyserver.

        This method is to be used for TESTING purposes only. The running
        gpgservice instance must have its test methods configured - something
        that should not be done in production. If this requirement is not met
        a RuntimeError will be raised.

        :param key: An IGPGKey instance.
        """
        document = {'keys': [{
            'owner': owner_id,
            'id': keyid,
            'fingerprint': fingerprint,
            'size': keysize,
            'algorithm': algorithm,
            'enabled': enabled,
            'can_encrypt': can_encrypt}]}
        path = '/test/add_keys'
        resp = self._request('post', path, json=document)
        if resp.status_code == http_codes['NOT_FOUND']:
            raise RuntimeError(
                "gpgservice was not configured with test endpoints enabled.")
        elif resp.status_code != http_codes['OK']:
            self.raise_for_error(resp)

    def getKeyDetailsFromKeyServer(self, fingerprint):
        """Get the details of a key from the configured key server.

        Retrieve a key from the key server identified by 'fingerprint'. Return
        the details of that key, or None if it could not be found.

        Note that this merely retrieves the key from the key server. It will
        not mutate the gpgservice database in any way.

        :raises ValueError: If the fingerprint isn't valid.
        :raises RuntimeError: If the gpgservice instance does not support the
            /keyserver/ API endpoint.
        :returns: A dictionary containing the key details.
        """
        fingerprint = sanitize_fingerprint_or_raise(fingerprint)
        path = '/keyserver/%s' % fingerprint
        resp = self._request('get', path)
        if resp.headers['content-type'] != 'application/json' and \
           resp.status_code == 404:
            raise RuntimeError(
                "The gpgservice instance running at %s does not support the "
                "'/keyserver' API endpoint." % self.get_endpoint())
        if resp.status_code == http_codes['NOT_FOUND']:
            return None
        else:
            return resp.json()

    def encryptContent(self, fingerprint, content):
        """Encrypt 'content' using the GPG key represented by 'fingerprint'"""
        fingerprint = sanitize_fingerprint_or_raise(fingerprint)
        resp = self._request(
            'post',
            '/encrypt',
            data=content,
            headers={'X-GPG-Fingerprint': fingerprint}
            )
        if resp.status_code != http_codes['OK']:
            self.raise_for_error(resp)
        return resp.content

    def verifySignedContent(self, content, signature=None):
        """Verify some signed content.

        :param content: The signed content you wish to verify.
        :param signature: If the signed content has a detached signature, this
            should be set to the signature of the document. If an inline
            signature is being used, this should remain at it's default value.
        :returns: A tuple of (fingerprint, content). 'fingerprint' is the
            fingerprint of the GPG key used to sign the content. 'content' is
            the content that was actually signed.
        :raises GPGServiceException: For error conditions, including missing
            keys, corrupted documents, multiple signatures and no signatures.
        """
        files = {
            'content': StringIO(content)
        }
        if signature is not None:
            files['signature'] = (
                'sig', StringIO(signature), 'application/pgp-signature')
        resp = self._request('post', '/verify', files=files)
        if resp.status_code == 200:
            return resp.headers['X-GPG-Verified-Fingerprint'], resp.content
        self.raise_for_error(resp)

    def _notify_writes(self):
        errors = []
        for hook in self.write_hooks:
            try:
                hook()
            except Exception as e:
                errors.append(str(e))
        if errors:
            raise Exception("The operation succeeded, but one or more write"
                            " hooks failed: %s" % ', '.join(errors))

    def _encode_owner_id(self, owner_id):
        return base64.b64encode(owner_id, altchars='-_')

    def raise_for_error(self, response):
        """Raise GPGServiceException based on what's in 'response'."""
        if response.headers['Content-Type'] == 'application/json':
            message = response.json()['status']
        else:
            message = (
                "Unhandled service error. HTTP Status: %d HTTP Body: %s" % (
                    response.status_code, response.content))
        raise GPGServiceException(message)

    def _request(self, method, path, data=None, json=None, **kwargs):
        assert data is None or json is None, "Specify json or data, not both"
        headers = kwargs.get('headers', dict())
        if json is not None:
            headers['Content-Type'] = 'application/json'
            data = json_dumps(json)
            kwargs['headers'] = headers
        response = getattr(requests, method)(
            urljoin(self.get_endpoint(), path), data=data,
            timeout=self.get_timeout(), **kwargs)
        return response
