# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import absolute_import

from ConfigParser import SafeConfigParser
import json
import os
import requests
import socket
from StringIO import StringIO
import subprocess
import sys
from tempfile import (
    mkstemp,
    NamedTemporaryFile,
)
from testtools.content import content_from_file
from textwrap import dedent
import time

from fixtures import Fixture


class GPGKeyServiceFixture(Fixture):
    """Run the gpgservice webapp and test key server."""

    def __init__(self, keyserver_host=None, keyserver_port=None,
                 gunicorn_path=None, service_start_timeout=5.0):
        """Create a new GPGKeyServiceFixture.

        :param config_fixture: If specified, this must be a ConfigFixture
            instance. It will be updated with the relevant GPG service config
            details.
        :param keyserver_host: Hostname for keyserver.
        :param keyserver_port: Port for keyserver.
        :param gunicorn_path: Path to gunicorn binary.
        :param service_start_timeout: The maximum time (in seconds) to wait
            for the gpgservice instance to start responding to HTTP queries.
        """

        self.keyserver_host = keyserver_host
        self.keyserver_port = keyserver_port
        self.gunicorn_path = gunicorn_path
        self.service_start_timeout = service_start_timeout

    def setUp(self):
        super(GPGKeyServiceFixture, self).setUp()
        # Write service config to a file on disk. This file gets
        # deleted when the fixture ends.
        service_config = _get_default_service_config(
            self.keyserver_host, self.keyserver_port)
        self._config_file = NamedTemporaryFile()
        self.addCleanup(self._config_file.close)
        service_config.write(self._config_file)
        self._config_file.flush()

        # Set the environment variable that tells gpgservice where to read its
        # config file from:
        env = os.environ.copy()
        env['GPGSERVICE_CONFIG_PATH'] = self._config_file.name

        if self.gunicorn_path is None:
            self.gunicorn_path = os.path.join('env', 'bin', 'gunicorn')
        self.interface = '127.0.0.1'
        self.port = _get_unused_port()
        gunicorn_options = ['-b', self.bind_address]
        wsgi_app_name = 'gpgservice.webapp:app'

        self._stdout_fd, self._stdout_fname = mkstemp()
        self._stderr_fd, self._stderr_fname = mkstemp()
        self._process = subprocess.Popen(
            args=[self.gunicorn_path] + gunicorn_options + [wsgi_app_name],
            stdout=self._stdout_fd,
            stderr=self._stderr_fd,
            env=env)
        self.addDetail('process-stdout', content_from_file(self._stdout_fname))
        self.addDetail('process-stderr', content_from_file(self._stderr_fname))
        self.addCleanup(os.remove, self._stdout_fname)
        self.addCleanup(os.remove, self._stderr_fname)
        self.addCleanup(self._kill_server)
        try:
            self._wait_for_service_start()
        except ErrorSummaryException:
            reraise_with_message(self._get_process_output_summary())
        self.reset_service_database()

    def _kill_server(self):
        self._process.terminate()

    def _wait_for_service_start(self):
        error_tracker = ExceptionTracker()
        while error_tracker.elapsed_seconds < self.service_start_timeout:
            try:
                resp = requests.get(self.root_url + '/')
            except (requests.exceptions.ConnectionError, socket.error) as e:
                error_tracker.add_exception(e)
            else:
                if resp.status_code == 200:
                    return
            time.sleep(0.1)
        error_tracker.raise_summary()

    def reset_service_database(self):
        """Reset gpgservice instance database to the launchpad sampledata."""
        test_data = {
            'keys': [
                {
                    'owner': 'name16_oid',
                    'id': '12345678',
                    'fingerprint': 'ABCDEF0123456789ABCDDCBA0000111112345678',
                    'size': 1024,
                    'algorithm': 'D',
                    'can_encrypt': True,
                    'enabled': True,
                }
            ]
        }
        headers = {'Content-Type': 'application/json'}
        resp = requests.post(
            self.root_url + '/test/reset_db',
            data=json.dumps(test_data),
            headers=headers,
            )
        if resp.status_code != 200:
            raise RuntimeError("Could not reset database: %s" % resp.content)

    def _get_process_output_summary(self):
        details = self.getDetails()
        return '\n'.join([
            'process-stdout:',
            details['process-stdout'].as_text(),
            'process-stderr:',
            details['process-stderr'].as_text(),
        ])

    @property
    def bind_address(self):
        return '%s:%d' % (self.interface, self.port)

    @property
    def root_url(self):
        return 'http://%s' % self.bind_address


def _get_default_service_config(keyserver_host, keyserver_port):
    config = SafeConfigParser()
    config.readfp(StringIO(dedent("""\
        [gpghandler]
        host: {host}
        public_host: keyserver.ubuntu.com
        upload_keys: True
        port: {port}
        timeout: 5.0
        maximum_upload_size: 16777216
        enable_test_endpoint: true

        [database]
        type: sqlite
    """.format(host=keyserver_host, port=keyserver_port))))
    return config


def _get_unused_port():
    """Find and return an unused port

    There is a small race condition here (between the time we allocate the
    port, and the time it actually gets used), but for the purposes for which
    this function gets used it isn't a problem in practice.
    """
    s = socket.socket()
    try:
        s.bind(('localhost', 0))
        return s.getsockname()[1]
    finally:
        s.close()


class ErrorSummaryException(Exception):
    """The exception raised from ExceptionTracker."""


class ExceptionTracker(object):
    """Keep track of repeated exception occurances.

    This class keeps track of the timeline of exceptions that occur while a
    certain operation is being performed. It should be instantiated as the
    operation begins. If the operation ends in an error, this class can raise
    an exception that contains the timeline of errors. A typical invocation
    pattern looks like this::

        errors = ExceptionTracker()
        for i in range(50):
            try:
                do_a_thing_that_might_fail()
                return
            except Exception as e:
                errors.add_exception(e)
                time.sleep(0.1)
        errors.raise_summary()

    The 'errors' object is 'truthy' if at least one error has been added, so
    you can also do things like this::

        if errors:
            # something went wrong!

    The ExceptionTracker class also keeps track of the time since it was
    created, so the above example could more elegantly be written as::

        errors = ExceptionTracker()
        while errors.elapsed_time < 5.0:
            try:
                do_a_thing_that_might_fail()
                return
            except Exception as e:
                errors.add_exception(e)
                time.sleep(0.1)
        errors.raise_summary()
    """

    def __init__(self):
        self.errors = []
        self.start_timestamp = time.time()

    def add_exception(self, exception):
        """Add an occurance of 'exception'.

        :raises TypeError: if 'exception' is not an instance of Exception.
        """
        if not isinstance(exception, Exception):
            raise TypeError(
                "'exception' must be an instance of Exception, not %s" %
                type(exception))
        if not self.errors or not self.errors[-1].is_same_exception(exception):
            self.errors.append(
                _ExceptionContainer(exception, self.elapsed_seconds))
        else:
            self.errors[-1].count += 1

    def error_summary(self):
        """Get a description of the errors raised so far.

        :returns: a multi-line string describing all the errors raised so far,
            or an empty string if no errors have been raised yet.
        """
        if not self.errors:
            return ''
        lines = [str(c) for c in self.errors]
        lines.append('Total elapsed time: %.3f Seconds' % self.elapsed_seconds)
        return '\n'.join(lines)

    def raise_summary(self):
        """Raise an exception that contains the summary of errors that occured.

        If no errors occured, this is a no-op.
        """
        if self.errors:
            raise ErrorSummaryException('\n' + self.error_summary())

    @property
    def elapsed_seconds(self):
        return time.time() - self.start_timestamp

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.errors)


class _ExceptionContainer(object):

    def __init__(self, exception, start_time_offset):
        self.exception = exception
        self.start_time_offset = start_time_offset
        self.count = 1

    def is_same_exception(self, exception):
        return isinstance(exception, type(self.exception)) and \
            str(self.exception) == str(exception)

    def __str__(self):
        return '+%.3f\t%d Errors: %s: %s' % (
            self.start_time_offset,
            self.count,
            type(self.exception).__name__,
            self.exception,
        )


def reraise_with_message(message):
    """Re-raise an existing exception, appending 'message'.

    This function annotates an existing raised exception with more information.
    A simple use of this function is as follows::

        try:
            do_some_setup()
        except Exception as e:
            details_message = get_system_state_details()
            reraise_with_message(details_message)

    If 'message' is an empty string, this function simply re-raises the
    original exception unchanged.

    :param message: A string you want to append to the original exception
        message.
    :raises ValueError: If this function is called while not processing an
        exception.
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    if exc_type is None:
        raise ValueError(
            "reraise_with_message must be called while handling an exception.")

    if not message:
        raise

    exc_message = ''.join((str(exc_value), '\n', message))
    raise exc_type, exc_type(exc_message), exc_traceback
