# Copyright 2009-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os
import shutil
import tempfile

from txfixtures.tachandler import TacTestFixture

from gpgservice_client.testing._fixture import _get_unused_port


__metaclass__ = type
KEYS_DIR = os.path.join(os.path.dirname(__file__), 'tests/keys')


class TestKeyServerFixture(TacTestFixture):
    """A test key server for use by functional tests."""

    def __init__(self, listen_host='localhost', listen_port=None,
                 working_dir=None):
        """Create a new TestKeyServerFixture.

        :param listen_host: The host to listen on. The default ('localhost')
            will cause the keyserver to bind to the loopback interface.
        :param listen_port: The TCP port to listen on, if None a randomised
            unused port will be provided.
        :param working_dir: The directory to write working files to. The
            default value of None will cause this fixture to create a new
            temporary directory for the lifetime of the fixture.
        """
        self._listen_host = listen_host
        self._listen_port = listen_port
        self._specified_working_dir = working_dir
        super(TestKeyServerFixture, self).__init__()

    # NOTE: Why have both setUp() and _setUp() ? I'm glad you asked...
    # testtools changed the way setUp works. The new best practise is to
    # override _setUp, rather then setUp. However, txfixtures hasn't been
    # updated to this new best practise. In addition, you cannot call
    # addCleanup until Fixture.setUp has been called.  All this means that we
    # have a rather complicated setup procedure:
    # - testtools.useFixture calls TestKeyServerFixture.setUp (see below)
    #  - we create the working directory if required, and set self._working_dir
    #    which determines what self.root, self.pidfile etc return.
    # - we call TacTestFixture.setUp.
    # - TacTestFixture.setUp calls Fixture.setUp
    # - Fixture.setUp enables the use of Fixture.addCleanup
    # - Fixture.setUp then calls _setUp, where we add a cleanup to remove the
    #   working directory. Note that this means it's the *first* cleanup on the
    #   stack (this will be important later)
    # - Fixture.setUp ends, and TacTestFixture.setUp continues to run. This is
    #   where twistd is started, and the killTac method is pushed onto the
    #   cleanup stack.
    #
    # - Once the fixture has ended, the cleanups are run in *reversed order*,
    #   so killTac is called *before* the working tree is removed.
    #
    # So, now you know...
    def setUp(self):
        """Check the environ for the twistd path."""
        twistd_path = os.environ.get('TWISTD_PATH')
        self._working_dir = self._specified_working_dir or tempfile.mkdtemp()
        if self._listen_port is None:
            self._listen_port = _get_unused_port()
        super(TestKeyServerFixture, self).setUp(twistd_script=twistd_path)

    def _setUp(self):
        super(TestKeyServerFixture, self)._setUp()
        if self._specified_working_dir is None:
            self.addCleanup(shutil.rmtree, self._working_dir)

    def setUpRoot(self):
        """Recreate root directory and copy needed keys"""
        if os.path.isdir(self.root):
            shutil.rmtree(self.root)
        shutil.copytree(KEYS_DIR, self.root)

    @property
    def daemon_port(self):
        return self._listen_port

    @property
    def bind_host(self):
        return self._listen_host

    @property
    def root(self):
        return self._working_dir

    @property
    def tacfile(self):
        if getattr(self, '_tacfile_path', None) is None:
            self._tacfile_path = tempfile.mktemp(suffix='.tac')
            with open(self._tacfile_path, 'w') as tacfile:
                tacfile.write('''
from twisted.application import (
    service,
    strports,
    )
from twisted.web import server

from gpgservice_client.testing.keyserver.web import KeyServerResource

application = service.Application('testkeyserver')
svc = service.IServiceCollection(application)

site = server.Site(KeyServerResource({working_dir!r}))
site.displayTracebacks = False

# Run on the port that gpghandler is configured to hit.
port = 'tcp:%s' % ({listen_port!r},)
strports.service(port, site).setServiceParent(svc)
                '''.format(
                    working_dir=self.root, listen_port=self.daemon_port))
            self.addCleanup(os.unlink, self._tacfile_path)
        return self._tacfile_path

    @property
    def pidfile(self):
        return os.path.join(self.root, 'testkeyserver.pid')

    @property
    def logfile(self):
        return os.path.join(self.root, 'testkeyserver.log')

    @property
    def url(self):
        """The URL that the web server will be running on."""
        return 'http://%s:%d' % (self._listen_host, self._listen_port)
