# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import time

from testtools import TestCase
from testtools.matchers import (
    Contains,
    Equals,
    GreaterThan,
    HasLength,
    MatchesRegex,
    Not,
    NotEquals,
    Raises,
)

from gpgservice_client.testing._fixture import (
    _ExceptionContainer,
    ErrorSummaryException,
    ExceptionTracker,
    GPGKeyServiceFixture,
    reraise_with_message,
)


class ServiceFixtureTests(TestCase):

    def test_collapses_service_start_errors(self):
        # Tell the fixture that gunicorn is at /bin/true.  This will cause it
        # to start the process successfully, but the HTTP queries will never
        # complete.
        fixture = GPGKeyServiceFixture(gunicorn_path='/bin/true')
        self.assertRaises(ErrorSummaryException, fixture.setUp)

    def test_attached_stdout_and_stderr_logs(self):
        fixture = GPGKeyServiceFixture()
        with fixture:
            # Details need to be gathered and examined *before* the fixture is
            # torn down - since at that point the file is deleted from disk.
            details = fixture.getDetails()
            self.assertThat(details, Contains('process-stdout'))
            self.assertThat(details, Contains('process-stderr'))
            # App logs to stderr on startup, stdout should be empty:
            self.assertThat(details['process-stdout'].as_text(), Equals(''))
            self.assertThat(details['process-stderr'].as_text(), NotEquals(''))


class ExceptionTrackerTests(TestCase):

    def test_empty_tracker_is_falsy(self):
        self.assertFalse(ExceptionTracker())

    def test_full_tracker_is_truthy(self):
        tracker = ExceptionTracker()
        tracker.add_exception(AssertionError("bork bork!"))
        self.assertTrue(tracker)

    def test_add_exception_raises_TypeError(self):
        tracker = ExceptionTracker()
        e = self.assertRaises(
            TypeError, tracker.add_exception, "Not an exception")
        self.assertEqual(
            "'exception' must be an instance of Exception, not <type 'str'>",
            str(e))

    def test_error_summary_returns_empty_string_on_empty_container(self):
        self.assertEqual('', ExceptionTracker().error_summary())

    def test_error_summary_with_one_error(self):
        tracker = ExceptionTracker()
        tracker.add_exception(AssertionError("Something went wrong!"))
        lines = tracker.error_summary().split('\n')

        self.assertThat(lines, HasLength(2))
        self.assertThat(
            lines[0],
            MatchesErrorLine(1, AssertionError, "Something went wrong!")
        )
        self.assertThat(lines[1], MatchesTimeElapsedLine())

    def test_error_summary_with_two_duplicate_errors(self):
        tracker = ExceptionTracker()
        tracker.add_exception(AssertionError("Something went wrong!"))
        tracker.add_exception(AssertionError("Something went wrong!"))
        lines = tracker.error_summary().split('\n')

        self.assertThat(lines, HasLength(2))
        self.assertThat(
            lines[0],
            MatchesErrorLine(2, AssertionError, "Something went wrong!")
        )
        self.assertThat(lines[1], MatchesTimeElapsedLine())

    def test_error_summary_with_two_different_errors(self):
        tracker = ExceptionTracker()
        tracker.add_exception(AssertionError("Something went wrong!"))
        tracker.add_exception(ValueError("Wrong value."))
        lines = tracker.error_summary().split('\n')

        self.assertThat(lines, HasLength(3))
        self.assertThat(
            lines[0],
            MatchesErrorLine(1, AssertionError, "Something went wrong!")
        )
        self.assertThat(
            lines[1],
            MatchesErrorLine(1, ValueError, "Wrong value.")
        )
        self.assertThat(lines[2], MatchesTimeElapsedLine())

    def test_error_summary_with_multiple_repeated_errors(self):
        tracker = ExceptionTracker()
        # Test the complex case: Multiple errors, followed by a different
        # error, followed by multiple errors of the first type. Should
        # condense into three error lines plus the status line.
        tracker.add_exception(AssertionError("Something went wrong!"))
        tracker.add_exception(AssertionError("Something went wrong!"))
        tracker.add_exception(AssertionError("Something went wrong!"))
        tracker.add_exception(ValueError("Wrong value."))
        tracker.add_exception(AssertionError("Something went wrong!"))
        tracker.add_exception(AssertionError("Something went wrong!"))
        lines = tracker.error_summary().split('\n')

        self.assertThat(lines, HasLength(4))
        self.assertThat(
            lines[0],
            MatchesErrorLine(3, AssertionError, "Something went wrong!")
        )
        self.assertThat(
            lines[1],
            MatchesErrorLine(1, ValueError, "Wrong value.")
        )
        self.assertThat(
            lines[2],
            MatchesErrorLine(2, AssertionError, "Something went wrong!")
        )
        self.assertThat(lines[3], MatchesTimeElapsedLine())

    def test_elapsed_seconds(self):
        tracker = ExceptionTracker()
        # Account for innacuracies in timing by sleeping for 300 ms, and
        # asserting that at least 200 ms have passed:
        time.sleep(0.3)
        self.assertThat(tracker.elapsed_seconds, GreaterThan(0.2))

    def test_raise_summary_is_noop_with_no_errors(self):
        tracker = ExceptionTracker()
        self.assertThat(tracker.raise_summary, Not(Raises()))

    def test_raise_summary_raises_ErrorSummaryException(self):
        tracker = ExceptionTracker()
        tracker.add_exception(AssertionError("Precondition failed."))
        e = self.assertRaises(ErrorSummaryException, tracker.raise_summary)
        self.assertEqual('\n' + tracker.error_summary(), str(e))


# A regular expression that represents the 'N.NNN' time offset display format:
time_regex = r'\d\.\d{3}'


def MatchesErrorLine(num_errors, error_type, error_text):
    return MatchesRegex(
        r'^\+' +
        time_regex +
        '\t%d Errors: %s: %s$' % (num_errors, error_type.__name__, error_text)
    )


def MatchesTimeElapsedLine():
    return MatchesRegex('^Total elapsed time: ' + time_regex + ' Seconds$')


class ExceptionContainerTests(TestCase):

    def test_same_exception_same_message_matches(self):
        exception_one = AssertionError("Message")
        exception_two = AssertionError("Message")
        container = _ExceptionContainer(exception_one, 0.0)
        self.assertTrue(container.is_same_exception(exception_two))

    def test_same_exception_different_messages_dont_match(self):
        exception_one = AssertionError("Message")
        exception_two = AssertionError("Other message")
        container = _ExceptionContainer(exception_one, 0.0)
        self.assertFalse(container.is_same_exception(exception_two))

    def test_different_exception_same_messages_dont_match(self):
        exception_one = AssertionError("Message")
        exception_two = ValueError("Message")
        container = _ExceptionContainer(exception_one, 0.0)
        self.assertFalse(container.is_same_exception(exception_two))

    def test_different_exception_different_messages_dont_match(self):
        exception_one = AssertionError("Message")
        exception_two = ValueError("Other Message")
        container = _ExceptionContainer(exception_one, 0.0)
        self.assertFalse(container.is_same_exception(exception_two))

    def test_str_with_one_exception(self):
        exception_one = AssertionError("Message")
        container = _ExceptionContainer(exception_one, 0.0)
        self.assertEqual(
            "+0.000\t1 Errors: AssertionError: Message",
            str(container)
        )

    def test_str_with_multiple_exceptions(self):
        exception_one = TypeError("'foo' must not be None.")
        container = _ExceptionContainer(exception_one, 0.0)
        container.count = 10
        self.assertEqual(
            "+0.000\t10 Errors: TypeError: 'foo' must not be None.",
            str(container)
        )

    def test_str_with_time_offset(self):
        exception_one = ValueError("'bar' must not be negative.")
        container = _ExceptionContainer(exception_one, 1.23456)
        self.assertEqual(
            "+1.235\t1 Errors: ValueError: 'bar' must not be negative.",
            str(container)
        )


class ReraiseWithDetailsTests(TestCase):

    def get_reraised_exception(self, exception, message=''):
        """Helper for testing the reraise_with_message function.

        This method raises 'exception', catches it, then calls
        reraise_with_message with the 'message' argument, catches that
        exception, and returns it.
        """
        try:
            raise exception
        except Exception:
            try:
                reraise_with_message(message)
            except Exception as e:
                return e

    def test_raises_valueerror_when_not_processing_an_exception(self):
        self.assertRaises(ValueError, reraise_with_message, '')

    def test_raises_original_exception_if_message_is_empty(self):
        exception = RuntimeError("The Widget is broken!")
        reraised_caught_exception = self.get_reraised_exception(exception)
        self.assertIs(reraised_caught_exception, exception)

    def test_exception_text_contains_original_and_details(self):
        e = self.get_reraised_exception(
            RuntimeError("This is the original exception text."),
            "Some additional message goes here",
        )
        output = str(e)
        self.assertTrue(
            output.startswith("This is the original exception text.\n"))
        self.assertIn("Some additional message goes here", output)
