# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from collections import namedtuple
import random
import string

from testtools import TestCase
from testtools.matchers import (
    AnyMatch,
    ContainsDict,
    EndsWith,
    Equals,
    HasLength,
    raises,
    StartsWith,
    )
from testtools.matchers._basic import _BinaryComparison

from gpgservice_client.client import (
    GPGClient,
    GPGServiceException,
    )
from gpgservice_client.testing.gpg_utils import decrypt_content
from gpgservice_client.testing.keyserver.harness import TestKeyServerFixture
from gpgservice_client.testing.keyserver.tests.keys import (
    SampleDataDocuments,
    SampleDataKeys,
    )
from gpgservice_client.testing._fixture import GPGKeyServiceFixture


def getUniqueHexString(digits=None):
    """Return a unique hexadecimal string.

    :param digits: The number of digits in the string. 'None' means you
        don't care.
    :return: A hexadecimal string, with 'a'-'f' in lower case.
    """
    hex_number = '%x' % random.randint(1, 999999)
    if digits is not None:
        hex_number = hex_number.zfill(digits)
    return hex_number


def makeGPGKey():
    """Make a crappy GPG key for the purposes of testing."""
    key_id = getUniqueHexString(digits=8).upper()
    fingerprint = key_id + 'A' * 32
    key = {
        'keyid': key_id,
        'fingerprint': fingerprint,
        'keysize': 1024,
        'algorithm': GPGKeyAlgorithm.R.name,
        'enabled': True,
        'can_encrypt': False
        }
    return key


class FakeMethod:
    """Catch any function or method call, and record the fact.

    Use this for easy stubbing.  The call operator can return a fixed
    value, or raise a fixed exception object.

    This is useful when unit-testing code that does things you don't
    want to integration-test, e.g. because it wants to talk to remote
    systems.
    """

    def __init__(self, result=None, failure=None):
        """Set up a fake function or method.

        :param result: Value to return.
        :param failure: Exception to raise.
        """
        self.result = result
        self.failure = failure

        # A log of arguments for each call to this method.
        self.calls = []

    def __call__(self, *args, **kwargs):
        """Catch an invocation to the method.

        Increment `call_count`, and adds the arguments to `calls`.

        Accepts any and all parameters.  Raises the failure passed to
        the constructor, if any; otherwise, returns the result value
        passed to the constructor.
        """
        self.calls.append((args, kwargs))

        if self.failure is None:
            return self.result
        else:
            raise self.failure

    @property
    def call_count(self):
        return len(self.calls)

    def extract_args(self):
        """Return just the calls' positional-arguments tuples."""
        return [args for args, kwargs in self.calls]

    def extract_kwargs(self):
        """Return just the calls' keyword-arguments dicts."""
        return [kwargs for args, kwargs in self.calls]


class GPGKeyAlgorithm:
    """
    GPG Compliant Key Algorithms Types:

    1 : "R", # RSA
    16: "g", # ElGamal
    17: "D", # DSA
    20: "G", # ElGamal, compromised

    This class is used to mat between the integer values (used by pgpme), the
    'name' (which is a single character identifier) and the title. This was
    originally a lazr.enum class, but we've dropped the zope dependency. This
    class is made to mimic the API of the lazr enum, which is why it's a little
    odd.

    """
    Algorithm = namedtuple('Algorithm', ['name', 'title'])

    R = Algorithm('R', 'R')
    LITTLE_G = Algorithm('LITTLE_G', 'g')
    D = Algorithm('D', 'D')
    G = Algorithm('G', 'G')

    items = {
        1: R,
        'R': R,
        16: LITTLE_G,
        'LITTLE_G': LITTLE_G,
        17: D,
        'D': D,
        20: G,
        'G': G
    }


class GPGClientTests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.keyserver_fixture = TestKeyServerFixture()
        cls.keyserver_fixture.setUp()
        cls.service_fixture = GPGKeyServiceFixture(
            keyserver_host=cls.keyserver_fixture.bind_host,
            keyserver_port=cls.keyserver_fixture.daemon_port)
        cls.service_fixture.setUp()
        cls.gpg_client = GPGClient(
            endpoint="http://{}".format(cls.service_fixture.bind_address),
            timeout=30.0)
        cls.gpg_client.registerWriteHook(cls._write_hook)

    @classmethod
    def tearDownClass(cls):
        cls.gpg_client.unregisterWriteHook(cls._write_hook)
        cls.service_fixture.cleanUp()

    def setUp(self):
        super(GPGClientTests, self).setUp()
        if getattr(self, 'has_written', False):
            self.service_fixture.reset_service_database()
        self.has_written = False

    @classmethod
    def _write_hook(cls):
        cls.has_written = True

    def get_random_owner_id_string(self):
        """Get a random string that's representative of the owner id scheme."""
        candidates = string.ascii_lowercase + string.digits
        openid_id = ''.join((random.choice(candidates) for i in range(6)))
        return 'http://testopenid.dev/+id/' + openid_id

    def test_raises_TypeError(self):
        self.assertRaises(TypeError, GPGClient)

    def test_get_key_for_user_with_sampledata(self):
        client = self.gpg_client
        data = client.getKeysForOwner('name16_oid')
        self.assertThat(data, ContainsDict({'keys': HasLength(1)}))

    def test_get_key_for_unknown_user(self):
        client = self.gpg_client

        user = self.get_random_owner_id_string()
        data = client.getKeysForOwner(user)
        self.assertThat(data, ContainsDict({'keys': HasLength(0)}))

    def test_register_non_callable_raises_TypeError(self):
        client = self.gpg_client
        self.assertThat(
            lambda: client.registerWriteHook("not a callable"),
            raises(TypeError))

    def test_unregister_with_unregistered_hook_raises_ValueError(self):
        client = self.gpg_client
        self.assertThat(
            lambda: client.unregisterWriteHook("not registered"),
            raises(ValueError))

    def test_can_unregister_registered_write_hook(self):
        client = self.gpg_client
        hook = FakeMethod()
        client.registerWriteHook(hook)
        client.unregisterWriteHook(hook)

        self.assertThat(
            lambda: client.unregisterWriteHook(hook),
            raises(ValueError))

    def test_can_add_new_fingerprint_for_user(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key
        user = self.get_random_owner_id_string()
        key = client.addKeyForOwner(user, fingerprint)

        IsExpectedKey = ContainsDict({
            'fingerprint': FingerprintsEqual(fingerprint),
            'enabled': Equals(True)
        })
        self.assertThat(key, IsExpectedKey)
        data = client.getKeysForOwner(user)
        self.assertThat(data, ContainsDict({'keys': HasLength(1)}))
        keys = data['keys']
        self.assertThat(keys[0], IsExpectedKey)

    def test_adding_fingerprint_notifies_writes(self):
        client = self.gpg_client
        hook = FakeMethod()
        client.registerWriteHook(hook)
        self.addCleanup(client.unregisterWriteHook, hook)
        fingerprint = SampleDataKeys.regular_key
        user = self.get_random_owner_id_string()
        client.addKeyForOwner(user, fingerprint)

        self.assertThat(hook.call_count, Equals(1))

    def test_adding_invalid_fingerprint_raises_ValueError(self):
        client = self.gpg_client
        user = self.get_random_owner_id_string()
        self.assertThat(
            lambda: client.addKeyForOwner(user, ''),
            raises(ValueError("Invalid fingerprint: ''.")))

    def test_adding_duplicate_fingerprint_raises_GPGServiceException(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key
        user_one = self.get_random_owner_id_string()
        user_two = self.get_random_owner_id_string()
        client.addKeyForOwner(user_one, fingerprint)
        self.assertThat(
            lambda: client.addKeyForOwner(user_two, fingerprint),
            raises(
                GPGServiceException(
                    "Error: Fingerprint already in database.")))

    def test_disabling_active_key(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key
        user = self.get_random_owner_id_string()
        client.addKeyForOwner(user, fingerprint)
        client.disableKeyForOwner(user, fingerprint)
        data = client.getKeysForOwner(user)

        self.assertThat(data, ContainsDict({'keys': HasLength(1)}))
        keys = data['keys']
        self.assertThat(keys[0], ContainsDict({'enabled': Equals(False)}))

    def test_disabling_key_notifies_writes(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key
        user = self.get_random_owner_id_string()
        client.addKeyForOwner(user, fingerprint)

        hook = FakeMethod()
        client.registerWriteHook(hook)
        self.addCleanup(client.unregisterWriteHook, hook)
        client.disableKeyForOwner(user, fingerprint)
        self.assertThat(hook.call_count, Equals(1))

    def test_disabling_invalid_fingerprint_raises_ValueError(self):
        client = self.gpg_client
        user = self.get_random_owner_id_string()
        self.assertThat(
            lambda: client.disableKeyForOwner(user, ''),
            raises(ValueError("Invalid fingerprint: ''."))
        )

    def test_can_get_key_by_fingerprint(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key
        user = self.get_random_owner_id_string()
        client.addKeyForOwner(user, fingerprint)

        key = client.getKeyByFingerprint(fingerprint)
        self.assertThat(
            key, ContainsDict({'owner': Equals(user),
                               'fingerprint': FingerprintsEqual(fingerprint)}))

    def test_get_missing_key_by_fingerprint(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key
        self.assertIsNone(client.getKeyByFingerprint(fingerprint))

    def test_get_key_with_bad_fingerprint_raises_ValueError(self):
        client = self.gpg_client
        self.assertThat(lambda: client.getKeyByFingerprint('bad fingerprint'),
                        raises(ValueError))

    def test_can_add_GPGKey_to_test_enabled_gpgservice(self):
        client = self.gpg_client
        gpgkey = makeGPGKey()
        user = self.get_random_owner_id_string()
        client.addKeyForTest(user, **gpgkey)

        key = client.getKeyByFingerprint(gpgkey['fingerprint'])

        self.assertThat(
            key, ContainsDict({'owner': Equals(user),
                               'fingerprint': Equals(gpgkey['fingerprint'])}))

    def test_can_retrieve_multiple_keys_by_fingerprints(self):
        client = self.gpg_client
        gpgkey_1 = makeGPGKey()
        gpgkey_2 = makeGPGKey()
        user = self.get_random_owner_id_string()
        client.addKeyForTest(user, **gpgkey_1)
        client.addKeyForTest(user, **gpgkey_2)

        data = client.getKeysByFingerprints(
            [gpgkey_1['fingerprint'], gpgkey_2['fingerprint']])

        self.assertThat(data, HasLength(2))
        self.assertThat(data, AnyMatch(
            ContainsDict(dict(fingerprint=Equals(gpgkey_1['fingerprint']))),
            ))
        self.assertThat(data, AnyMatch(
            ContainsDict(dict(fingerprint=Equals(gpgkey_2['fingerprint'])))
            ))

    def test_retrieve_nonexistant_key_from_keyserver_returns_None(self):
        client = self.gpg_client
        gpgkey_1 = makeGPGKey()

        result = client.getKeyDetailsFromKeyServer(gpgkey_1['fingerprint'])
        self.assertIsNone(result)

    def test_retrieving_sampledata_key_from_keyserver_works(self):
        client = self.gpg_client
        fingerprint = SampleDataKeys.regular_key

        result = client.getKeyDetailsFromKeyServer(fingerprint)
        # Ensure it gives us some details in a dict - the full response is
        # tested in gpgservice itself.
        self.assertThat(result, ContainsDict({'can_encrypt': Equals(True)}))

    def test_retrieve_key_from_keyserver_raises_on_invalid_key(self):
        self.assertThat(
            lambda: self.gpg_client.getKeyDetailsFromKeyServer('invalid'),
            raises(ValueError)
        )

    def test_can_encrypt_content(self):
        client = self.gpg_client
        plain_text = self.getUniqueString()
        encryption_fingerprint = SampleDataKeys.regular_key

        encrypted_data = client.encryptContent(
            encryption_fingerprint, plain_text)
        self.assertThat(
            encrypted_data, StartsWith('-----BEGIN PGP MESSAGE-----'))
        self.assertThat(
            encrypted_data, EndsWith('-----END PGP MESSAGE-----\n'))

    def test_encrypt_with_signing_key_errors(self):
        client = self.gpg_client
        plain_text = self.getUniqueString()
        encryption_fingerprint = SampleDataKeys.sign_only_key

        self.assertThat(
            lambda: client.encryptContent(encryption_fingerprint, plain_text),
            raises(GPGServiceException)
            )

    def test_encrypt_with_revoked_key_errors(self):
        client = self.gpg_client
        plain_text = self.getUniqueString()
        encryption_fingerprint = SampleDataKeys.revoked_key

        self.assertThat(
            lambda: client.encryptContent(encryption_fingerprint, plain_text),
            raises(GPGServiceException)
            )

    def test_encrypt_with_expired_key_errors(self):
        client = self.gpg_client
        plain_text = self.getUniqueString()
        encryption_fingerprint = SampleDataKeys.expired_key

        self.assertThat(
            lambda: client.encryptContent(encryption_fingerprint, plain_text),
            raises(GPGServiceException)
            )

    def test_utils_can_decrypt_encrypted_content(self):
        client = self.gpg_client
        plain_text = self.getUniqueString()
        encryption_fingerprint = SampleDataKeys.regular_key

        encrypted_data = client.encryptContent(
            encryption_fingerprint, plain_text)
        new_plaintext = decrypt_content(encrypted_data)

        self.assertEqual(plain_text, new_plaintext)

    def test_can_verify_signed_content(self):
        client = self.gpg_client
        fingerprint, content = client.verifySignedContent(
            SampleDataDocuments.clearsigned_document)

        self.assertThat(content, Equals("This is a test document.\n"))
        self.assertThat(
            fingerprint, FingerprintsEqual(SampleDataKeys.sign_only_key))

    def test_can_verify_signed_content_with_unsigned_parts(self):
        client = self.gpg_client
        fingerprint, content = client.verifySignedContent(
            SampleDataDocuments.clearsigned_with_unsigned_parts)

        self.assertThat(content, Equals("This is a test document.\n"))
        self.assertThat(
            fingerprint, FingerprintsEqual(SampleDataKeys.sign_only_key))

    def test_errors_with_multiple_signatures(self):
        self.assertRaises(
            GPGServiceException,
            self.gpg_client.verifySignedContent,
            SampleDataDocuments.clearsigned_document_multiple_signatures
            )

    def test_errors_with_no_signature(self):
        self.assertRaises(
            GPGServiceException,
            self.gpg_client.verifySignedContent,
            SampleDataDocuments.clearsigned_document_no_signature
            )

    def test_errors_with_unknown_key(self):
        self.assertRaises(
            GPGServiceException,
            self.gpg_client.verifySignedContent,
            SampleDataDocuments.clearsigned_with_unknown_key
            )


class FingerprintsEqual(_BinaryComparison):

    """A matcher that understands how to compare GPG fingerprints."""

    @staticmethod
    def compare_fingerprints(a, b):
        return a.replace(' ', '') == b.replace(' ', '')

    comparator = compare_fingerprints
    mismatch_string = '<spaceless-comparison>'
